import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "moebius92.RaceTrack";

_G.trim = function(s)
	return string.gsub(s, "^%s*(.-)%s*$", "%1");
end

quickKeys = { 0x1000002A, 0x10000032, 0x1000003E, 0x10000047, 0x10000050, 0x10000057,
              0x1000005F, 0x10000067, 0x10000071, 0x1000004F, 0x10000056, 0x1000005E,

              0x10000066, 0x10000070, 0x10000078, 0x10000082, 0x1000008E, 0x10000096,
              0x1000009E, 0x10000011, 0x10000019, 0x10000022, 0x10000029, 0x10000031,

              0x1000003D, 0x10000046, 0x1000004E, 0x10000055, 0x1000005D, 0x10000005,
              0x1000000B, 0x10000010, 0x10000018, 0x10000021, 0x10000028, 0x10000030,

              0x1000003C, 0x10000045, 0x1000004D, 0x1000008D, 0x10000095, 0x1000009D,
              0x100000A3, 0x100000AD, 0x100000B0, 0x100000B9, 0x10000004, 0x1000000A,

              0x1000000F, 0x10000020, 0x10000027, 0x1000002F, 0x1000003B, 0x10000044,
              0x1000004C, 0x10000054, 0x1000005C, 0x10000065, 0x1000006F, 0x100000AC,

              0x100000AF, 0x100000B8, 0x10000003, 0x10000009, 0x1000000E, 0x10000017,
              0x1000001F, 0x10000026, 0x1000002E, 0x1000009C, 0x100000A2, 0x100000AB };

recovery = { blinding = true, bane = true, cracked = false, herb = false, storm = true, march = true, light = true, test = true };

quickslot = { blinding = "0x70003152", bane = "0x7000B60D", cracked = "0x70003143", herb = "0x70003EC0",
              storm = "0x7000B63C", march = "0x70000FA7", light = "0x70003EC2", test = "0x70003EB9" };

ccName = {
	["Blinding Flash"] = "blinding",
	["Bane Flare"] = "bane",
	["Cracked Earth"] = "cracked",
	["Herb-lore"] = "herb",
	["Storm-lore"] = "storm",
	["Ents Go To War"] = "march",
	["Light of the Rising Dawn"] = "light",
	["Test of Will"] = "test"
};

CCTimerRegime = class(Turbine.UI.Control);

function CCTimerRegime:Constructor(msg)
	Turbine.UI.Control.Constructor(self);
	self.msg = msg;
	self:SetMouseVisible(false);
end

function CCTimerRegime:MouseEnter(args)
	Turbine.Shell.WriteLine(self.msg);
end

function CCTimerRegime:MouseLeave(args)
	Turbine.Shell.WriteLine("---");
end

CCTimerWindow = class(Turbine.UI.Window);

function CCTimerWindow:Constructor()
	Turbine.UI.Window.Constructor(self);

	self.settings = CCTimerWindowLoadSettings();

	self.hide = false;
	self.visible = false;
	self.size = 36;
	self.seconds = 30;
	self.dragging = false;
	self.move = false;
	self.lanes = 1;
	self:SetPosition(self.settings.x, self.settings.y);
	self:SetMouseVisible(false);
	self.lastTime = Turbine.Engine.GetGameTime();
	self.track = moebius92.RaceTrack.RaceTrack(self.size, 0);
	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.name = self.player:GetName();

	Turbine.PluginManager.LoadPlugin("SkillInfo");

	self.mark = { };
	for k, v in pairs(ccName) do
		self.mark[v] = 0;
	end

	self:SetBackColor(Turbine.UI.Color(0.25, 0.0, 1.0, 0.0));
	self:SetWantsKeyEvents(true);
	skillInfoWindow:AddQuickslotHandler(function(skill, time, type)
		local n = ccName[skill];

		if n ~= nil then
			if type == "P" then
				self.mark[n] = time;
			elseif type == "H" or type == "C" or type == "D" then
				if (time - self.mark[n]) < 2.0 then
					self:AddCC(n, self.settings.cc[n], self.mark[n]);
				end
				self.mark[n] = 0;
			end
		end
	end);
	self:Layout(true);
	self:SetVisible();
end

function CCTimerWindow:SetVisible(v)
	if v ~= nil then
		self.visible = v;
	end

	Turbine.UI.Window.SetVisible(self, self.visible and not self.hide);
end

function CCTimerWindow:IsVisible()
	return self.visible;
end

function CCTimerWindow:Layout(lanes)
	local w = self.size * self.lanes + 2;

	if not lanes then
		self:SetWidth(w);
		for k, v in pairs(self.regimes) do
			v:SetWidth(w);
		end
	else
		local h = self.size + self.seconds * 10 + 10 - 4;
		local m = math.max(self.settings.blinding, self.settings.test, self.settings.light);

		if self.regimes ~= nil then
			for k, v in pairs(self.regimes) do
				v:SetParent(nil);
			end
		end

		self.regimes = { };

		self:SetSize(w, h);

		self.regimes.recovery = CCTimerRegime("recovery state\non entry, it is safe to use fan the flames on stunned targets");
		self.regimes.recovery:SetParent(self);
		self.regimes.recovery:SetBackColor(Turbine.UI.Color(0.25, 1.0, 0.0, 0.0));
		self.regimes.recovery:SetSize(w, 10);
		self.regimes.recovery:SetPosition(0, h - 10);

		self.regimes.reroot = CCTimerRegime("on entry, it is safe to use cracked earth on rooted targets");
		self.regimes.reroot:SetParent(self);
		self.regimes.reroot:SetBackColor(Turbine.UI.Color(0.25, 0.0, 0.0, 1.0));
		self.regimes.reroot:SetSize(w, (self.settings.reroot - m) * 10);
		self.regimes.reroot:SetPosition(0, h - self.settings.reroot * 10 - 10);

		local s = { };
		local e = { };
		local i = { };

		s.blinding = { };
		s.blinding.start = self.settings.blinding * 10;
		s.blinding.stop = self.settings.blinding * 10 - 11;
		s.blinding.r = 1.0;
		s.blinding.g = 1.0;
		s.blinding.b = 0.0;

		s.light = { };
		s.light.start = self.settings.light * 10;
		s.light.stop = self.settings.light * 10 - 14;
		s.light.r = 1.0;
		s.light.g = 0.0;
		s.light.b = 0.0;

		s.test = { };
		s.test.start = self.settings.test * 10;
		s.test.stop = self.settings.test * 10 - 11;
		s.test.r = 1.0;
		s.test.g = 0.0;
		s.test.b = 0.0;

		for k, v in pairs(s) do
			e[v.stop] = { n = 0, r = 0, g = 0, b = 0 };
		end

		for k, v in pairs(s) do
			if e[v.start] == nil then
				e[v.start] = { n = 0, r = 0, g = 0, b = 0 };
			end
			e[v.start].n = e[v.start].n + 1;
			e[v.start].r = e[v.start].r + v.r;
			e[v.start].g = e[v.start].g + v.g;
			e[v.start].b = e[v.start].b + v.b;

			if e[v.stop] == nil then
				e[v.stop] = { n = 0, r = 0, g = 0, b = 0 };
			end

			e[v.stop].n = e[v.stop].n - 1;
			e[v.stop].r = e[v.stop].r - v.r;
			e[v.stop].g = e[v.stop].g - v.g;
			e[v.stop].b = e[v.stop].b - v.b;
		end

		for k, v in pairs(e) do
			table.insert(i, k);
		end
		table.sort(i, function (a, b) return a > b end);

		local r = 0;
		local g = 0;
		local b = 0;
		local n = 0;
		local rc, gc, bc, mc;

		for k, v in ipairs(i) do
			n = n + e[v].n;
			r = r + e[v].r;
			g = g + e[v].g;
			b = b + e[v].b;

			if n > 0 then
				msg = "";
				if v == s.light.start then
					msg = msg .. "on entry, it is not safe to use light of the rising dawn on dazed targets\n";
					msg = msg .. "on entry, it is safe to use bane flare on stunned targets\n";
				end
				if v == s.test.start then
					msg = msg .. "on entry, it is not safe to use test of will on dazed targets\n";
				end
				if v == s.blinding.start then
					msg = msg .. "on entry, it is not safe to use blinding flash on dazed targets\n";
				end

				if i[k + 1] == s.light.stop then
					msg = msg .. "on exit, it is safe to use light of the rising dawn on dazed/stunned targets\n";
				end
				if i[k + 1] == s.test.stop then
					msg = msg .. "on exit, it is safe to use test of will on dazed/stunned targets\n";
				end
				if i[k + 1] == s.blinding.stop then
					msg = msg .. "on exit, it is safe to use blinding flash on dazed/stunned targets\n";
				end

				self.regimes[v] = CCTimerRegime(msg);
				self.regimes[v]:SetParent(self);
				rc = r / n;
				gc = g / n;
				bc = b / n;
				mc = math.max(rc, gc, bc);
				if n > 1 then
					rc = 1.0 - rc / mc;
					gc = 1.0 - gc / mc;
					bc = 1.0 - bc / mc;
					mc = math.max(rc, gc, bc);
				end
				if mc > 0 then
					self.regimes[v]:SetBackColor(Turbine.UI.Color(0.25, rc / mc, gc / mc, bc / mc));
				else
					self.regimes[v]:SetBackColor(Turbine.UI.Color(0.25, 0.0, 0.0, 0.0));
				end
				self.regimes[v]:SetMouseVisible(false);
				self.regimes[v]:SetSize(w, v - i[k + 1]);
				self.regimes[v]:SetPosition(0, h - v - 10);
			end
		end
	end
end

function CCTimerWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left and self.move then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.dragging = true;
	end
end

function CCTimerWindow:MouseMove(args)
	if self.dragging then
		local x, y = self:GetPosition();
		x = x + (args.X - self.dragStartX);
		y = y + (args.Y - self.dragStartY);
		self.settings.x = x;
		self.settings.y = y;
		self:SetPosition(x, y);
	end
end

function CCTimerWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self:SaveSettings();
		self.dragging = false;
	end
end

function CCTimerWindow:UpdateLayout()
	local time = Turbine.Engine.GetGameTime();

	self.track:set(1);
	ptr = self.head;
	while ptr ~= nil do
		if ptr.startTime > time then
			ptr.lane = nil;
		else
			local y = math.floor((ptr.endTime - time) * 10);
			ptr.lane = self.track:place(-y);
		end
		ptr = ptr.next;
	end

	if self.lanes ~= self.track.lanes then
		self.lanes = self.track.lanes;
		self:Layout(false);
		self:SetVisible(false);
		self:SetVisible(true);
	end
end

function CCTimerWindow:Update()
	local ptr, prev, next;
	local time = Turbine.Engine.GetGameTime();
	local delta;
	local changed = false;

	ptr = self.head;
	prev = nil;
	while ptr ~= nil do
		delta = ptr.endTime - time;
		next = ptr.next;

		if delta < 0.0 and ptr.pre then
			changed = true;
			if prev == nil then
				self.head = ptr.next;
			else
				prev.next = ptr.next;
			end

			ptr.endTime = ptr.endTime + self.settings.cc[ptr.type].duration;
			ptr.pre = false;
			ptr:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, quickslot["herb"]));
			if self.head == nil or self.head.endTime >= ptr.endTime then
				ptr.next = self.head;
				self.head = ptr;
			else
				local p = self.head;
				while p.next ~= nil and p.next.endTime < ptr.endTime do
					p = p.next;
				end
				ptr.next = p.next;
				p.next = ptr;
				if p == prev then
					prev = ptr;
				end
			end
		else
			prev = ptr;
		end

		ptr = next;
	end

	ptr = self.head;
	prev = nil;
	while ptr ~= nil do
		delta = ptr.endTime - time;

		if delta < -1.0 or (delta < 0.0 and not recovery[ptr.type]) then
			changed = true;
			ptr:SetParent(nil);
			if prev == nil then
				self.head = ptr.next;
			else
				prev.next = ptr.next;
			end
		else
			prev = ptr;
		end

		ptr = ptr.next;
	end

	ptr = self.head;
	while ptr ~= nil do
		if ptr.startTime <= time and ptr.lane == nil then
			changed = true;
		end
		ptr = ptr.next;
	end

	if changed then
		self:UpdateLayout();
	end

	ptr = self.head;
	while ptr ~= nil do
		if ptr.lane == nil then
			ptr:SetPosition(0, -self.size);
		else
			local y = math.floor((ptr.endTime - time) * 10);
			ptr:SetPosition(self.size * ptr.lane, -3 + (self.seconds - (ptr.endTime - time)) * 10);
		end
		ptr = ptr.next;
	end

	if self.head == nil then
		self:SetWantsUpdates(false);
		self:SetVisible(false);
	end
end

function CCTimerWindow:AddCC(k, v, time)
	self:SetVisible(true);

	local cc = Turbine.UI.Lotro.Quickslot();

	if v.pre ~= nil then
		cc.endTime = time + v.pre + v.act;
		cc.pre = true;
	else
		cc.endTime = time + v.duration + v.act;
		cc.pre = false;
	end

	cc.type = k;
	cc.startTime = time + v.act;
	cc:SetParent(self);
	cc:SetSize(self.size, self.size);
	cc:SetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, quickslot[k]));
	cc:SetMouseVisible(false);
	cc:SetZOrder(1);

	if self.head == nil or self.head.endTime >= cc.endTime then
		cc.next = self.head;
		self.head = cc;
	else
		local ptr = self.head;
		while ptr.next ~= nil and ptr.next.endTime < cc.endTime do
			ptr = ptr.next;
		end
		cc.next = ptr.next;
		ptr.next = cc;
	end

	self.lastTime = time;
	self:UpdateLayout();
	self:Update();
	self:SetWantsUpdates(true);
end

function CCTimerWindow:KeyDown(args)
	local time = Turbine.Engine.GetGameTime();

	if args.Action == 0x1000007B then
		if self.move then
			self.move = false;
			if self.head == nil then
				self:SetVisible(false);
			end
			if self.dragging then
				self:SaveSettings();
				self.dragging = false;
			end
		else
			self.move = true;
			self:SetVisible(true);
		end

		self:SetMouseVisible(self.move);
		for k, v in pairs(self.regimes) do
			v:SetMouseVisible(self.move);
		end
	elseif args.Action == 0x100000B3 then
		self.hide = not self.hide;
		self:SetVisible();
	end
end

function CCTimerWindow:SaveSettings()
	Turbine.PluginData.Save(Turbine.DataScope.Character, "CCTimer", self.settings);
end

function CCTimerWindowLoadSettings()
	settings = Turbine.PluginData.Load(Turbine.DataScope.Character, "CCTimer");

	if type(settings) ~= "table" then
		settings = { };
	end

	if not settings.cc then
		settings.cc = { };
	end

	if not settings.cc.blinding then
		settings.cc.blinding = { };
		settings.cc.blinding.duration = 30;
		settings.cc.blinding.act = 1.1;
	end

	if not settings.cc.bane then
		settings.cc.bane = { };
		settings.cc.bane.duration = 15;
		settings.cc.bane.act = 1.1;
	end

	if not settings.cc.light then
		settings.cc.light = { };
		settings.cc.light.duration = 3;
		settings.cc.light.act = 0.2;
	end

	if not settings.cc.test then
		settings.cc.test = { };
		settings.cc.test.duration = 5;
		settings.cc.test.act = 0.5;
	end

	if not settings.cc.herb then
		settings.cc.herb = { };
		settings.cc.herb.duration = 30;
		settings.cc.herb.act = -0.2;
	end

	if not settings.cc.storm then
		settings.cc.storm = { };
		settings.cc.storm.duration = 3;
		settings.cc.storm.act = -0.2;
	end

	if not settings.cc.cracked then
		settings.cc.cracked = { };
		settings.cc.cracked.pre = 5;
		settings.cc.cracked.duration = 30;
		settings.cc.cracked.act = 0.2;
	end

	if not settings.cc.march then
		settings.cc.march = { };
		settings.cc.march.duration = 6;
		settings.cc.march.act = 1.1;
	end

	if not settings.x then
		settings.x = 0;
	end

	if not settings.y then
		settings.y = 0;
	end

	if not settings.blinding then
		settings.blinding = 1.5;
	end

	if not settings.light then
		settings.light = 1.8;
	end

	if not settings.test then
		settings.test = 3.1;
	end

	if not settings.reroot then
		settings.reroot = 7.3;
	end

	return settings;
end

CCTimerCommand = class(Turbine.ShellCommand);

function CCTimerCommand:Execute(command, arguments)
	local i = string.find(arguments, " ");
	local arg, cc, rest;

	if i ~= nil then
		arg = trim(string.sub(arguments, 0, i - 1));
		rest = trim(string.sub(arguments, i + 1, -1));
		if rest == "" then
			rest = nil;
		end
	else
		arg = arguments;
		rest = nil;
	end

	if arg == "timer" then
		local time = Turbine.Engine.GetGameTime();
		Turbine.Shell.WriteLine((time - ccTimerWindow.lastTime) .. " seconds.");
	elseif arg == "reroot" then
		i = tonumber(rest);

		if i ~= nil then
			ccTimerWindow.settings.reroot = i;
			ccTimerWindow:Layout(true);
			ccTimerWindow:SaveSettings();
		end

		Turbine.Shell.WriteLine("CCTimer reroot time is " .. ccTimerWindow.settings.reroot);
	elseif arg == "blinding" then
		i = tonumber(rest);

		if i ~= nil then
			ccTimerWindow.settings.blinding = i;
			ccTimerWindow:Layout(true);
			ccTimerWindow:SaveSettings();
		end

		Turbine.Shell.WriteLine("Blinding Flash unsafe time is " .. ccTimerWindow.settings.blinding);
	elseif arg == "light" then
		i = tonumber(rest);

		if i ~= nil then
			ccTimerWindow.settings.light = i;
			ccTimerWindow:Layout(true);
			ccTimerWindow:SaveSettings();
		end

		Turbine.Shell.WriteLine("Light of the Rising Dawn unsafe time is " .. ccTimerWindow.settings.light);
	elseif arg == "test" then
		i = tonumber(rest);

		if i ~= nil then
			ccTimerWindow.settings.test = i;
			ccTimerWindow:Layout(true);
			ccTimerWindow:SaveSettings();
		end

		Turbine.Shell.WriteLine("Test of Wills unsafe time is " .. ccTimerWindow.settings.test);
	elseif arg == "act" or arg == "duration" or arg == "pre" then
		i = string.find(rest, " ");
		if i ~= nil then
			cc = trim(string.sub(rest, 0, i - 1));
			rest = trim(string.sub(rest, i + 1, -1));
			i = tonumber(rest);
		else
			cc = rest;
		end

		if i ~= nil then
			if i == 0 and arg == "pre" then
				i = nil;
			end
			ccTimerWindow.settings.cc[cc][arg] = i;
			ccTimerWindow:SaveSettings();
		end
		if ccTimerWindow.settings.cc[cc][arg] ~= nil then
			Turbine.Shell.WriteLine("CCTimer " .. cc .. "(" .. arg .. ") is " .. ccTimerWindow.settings.cc[cc][arg]);
		else
			Turbine.Shell.WriteLine("CCTimer " .. cc .. "(" .. arg .. ") is 0");
		end
	else
		Turbine.Shell.WriteLine("usage: /cctimer <<act|duration|pre> <cc> <#>|reroot <#>|<blinding|light|test> <#>|timer>");
	end
end

function CCTimerCommand:GetShortHelp()
	return "Loremaster Crowd Control Timer";
end

ccTimerCommand = CCTimerCommand();
Turbine.Shell.AddCommand('cctimer', ccTimerCommand);
Turbine.Shell.AddCommand('cc', ccTimerCommand);
ccTimerWindow = CCTimerWindow();